use tauri::App;
use tauri::AppHandle;
use tauri::Emitter;
use tauri::Manager;
use tauri::Runtime;
use tauri_plugin_global_shortcut::GlobalShortcutExt;
use tauri_plugin_global_shortcut::Shortcut;
use tauri_plugin_global_shortcut::ShortcutState;
use tauri_plugin_store::JsonValue;
use tauri_plugin_store::StoreExt;

/// Name of the Tauri storage
const HAYSTACK_TAURI_STORE: &str = "haystack_tauri_store";

/// Key for storing global shortcuts
const HAYSTACK_GLOBAL_SHORTCUT: &str = "haystack_global_shortcut";

/// Default shortcut for macOS
#[cfg(target_os = "macos")]
const DEFAULT_SHORTCUT: &str = "command+shift+k";

/// Default shortcut for Windows and Linux
#[cfg(any(target_os = "windows", target_os = "linux"))]
const DEFAULT_SHORTCUT: &str = "ctrl+shift+k";

/// Set shortcut during application startup
pub fn enable_shortcut(app: &App) {
    let store = app
        .store(HAYSTACK_TAURI_STORE)
        .expect("Creating the store should not fail");

    let shortcut = if let Some(stored_shortcut) = store.get(HAYSTACK_GLOBAL_SHORTCUT) {
        let stored_shortcut_str = match stored_shortcut {
            JsonValue::String(str) => str,
            unexpected_type => panic!(
                "Haystack shortcuts should be stored as strings, found type: {} ",
                unexpected_type
            ),
        };
        stored_shortcut_str
            .parse::<Shortcut>()
            .expect("Stored shortcut string should be valid")
    } else {
        store.set(
            HAYSTACK_GLOBAL_SHORTCUT,
            JsonValue::String(DEFAULT_SHORTCUT.to_string()),
        );
        DEFAULT_SHORTCUT
            .parse::<Shortcut>()
            .expect("Default shortcut should be valid")
    };

    register_shortcut_upon_start(app, shortcut);
}

/// Get the current stored shortcut as a string
#[tauri::command]
pub fn get_current_shortcut<R: Runtime>(app: AppHandle<R>) -> Result<String, String> {
    Ok(get_shortcut_from_store(&app))
}

/// Unregister the current shortcut in Tauri
#[tauri::command]
pub fn unregister_shortcut<R: Runtime>(app: AppHandle<R>) {
    let shortcut_str = get_shortcut_from_store(&app);
    let shortcut = shortcut_str
        .parse::<Shortcut>()
        .expect("Stored shortcut string should be valid");

    app.global_shortcut()
        .unregister(shortcut)
        .expect("Failed to unregister shortcut")
}

/// Change the global shortcut
#[tauri::command]
pub fn change_shortcut<R: Runtime>(
    app: AppHandle<R>,
    _window: tauri::Window<R>,
    key: String,
) -> Result<(), String> {
    println!("Key: {}", key);
    let shortcut = match key.parse::<Shortcut>() {
        Ok(shortcut) => shortcut,
        Err(_) => return Err(format!("Invalid shortcut {}", key)),
    };

    // Store the new shortcut
    let store = app
        .get_store(HAYSTACK_TAURI_STORE)
        .expect("Store should already be loaded or created");
    store.set(HAYSTACK_GLOBAL_SHORTCUT, JsonValue::String(key));

    // Register the new shortcut
    register_shortcut(&app, shortcut);

    Ok(())
}

/// Helper function to handle window visibility toggle
fn handle_window_visibility<R: Runtime>(app: &AppHandle<R>, window: &tauri::WebviewWindow<R>) {
    if window.is_visible().unwrap() {
        window.hide().unwrap();
    } else {
        window.show().unwrap();
        window.set_focus().unwrap();
        app.emit("focus-search", ()).unwrap();
    }
}

/// Helper function to register a shortcut, primarily for updating shortcuts
fn register_shortcut<R: Runtime>(app: &AppHandle<R>, shortcut: Shortcut) {
    let main_window = app.get_webview_window("main").unwrap();

    app.global_shortcut()
        .on_shortcut(shortcut, move |app, scut, event| {
            if scut == &shortcut {
                if let ShortcutState::Pressed = event.state() {
                    handle_window_visibility(app, &main_window);
                }
            }
        })
        .map_err(|err| format!("Failed to register new shortcut '{}'", err))
        .unwrap();
}

/// Helper function to register shortcuts during application startup
fn register_shortcut_upon_start(app: &App, shortcut: Shortcut) {
    let window = app
        .get_webview_window("main")
        .expect("webview to be defined");

    app.handle()
        .plugin(
            tauri_plugin_global_shortcut::Builder::new()
                .with_handler(move |app, scut, event| {
                    if scut == &shortcut {
                        if let ShortcutState::Pressed = event.state() {
                            handle_window_visibility(app, &window);
                        }
                    }
                })
                .build(),
        )
        .unwrap();
    app.global_shortcut().register(shortcut).unwrap();
}

/// Retrieve the stored global shortcut as a string
fn get_shortcut_from_store<R: Runtime>(app: &AppHandle<R>) -> String {
    let store = app
        .get_store(HAYSTACK_TAURI_STORE)
        .expect("Store should already be loaded or created");

    match store
        .get(HAYSTACK_GLOBAL_SHORTCUT)
        .expect("Shortcut should already be stored")
    {
        JsonValue::String(str) => str,
        unexpected_type => panic!(
            "Haystack shortcuts should be stored as strings, found type: {} ",
            unexpected_type
        ),
    }
}
